import 'dart:convert';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare/src/modules/logger/logger.dart';
import 'package:contentsquare/src/modules/screengraph/compression.dart';
import 'package:contentsquare/src/modules/screengraph/render_object_extension.dart';
import 'package:contentsquare/src/modules/webview_tracking/javascript_event_handler.dart';
import 'package:fast_immutable_collections/fast_immutable_collections.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';

import '../screengraph/hit_test/get_element_data_of_hit_render_object.dart';
import '../screengraph/hit_test/hit_element_data.dart';
import 'exceptions/exceptions.dart';
import 'logs/logs.dart';
import 'screengraph_retrieval.dart';

/// Widget used to track a WebView.
///
/// {@template contentsquare.webview.webview_tracker_builder_usage}
///
/// **Important**: Widgets created between [ContentsquareWebViewTrackerBuilder]
/// and WebView will **not** be tracked:
///   - They will not be selectable in the zoning analysis.
///   - Gestures on those widgets will not be tracked.
/// Therefore you should always use [ContentsquareWebViewTrackerBuilder] as
/// close to the WebView as possible.
/// {@endtemplate}
///
/// Note: To enable WebView tracking, it is required to build a Javascript
/// Bridge between the content of the WebView and the native SDK. To do so,
/// you will have to implement the Contentsquare WebView Javascript Tracking
/// Tag in the web pages called in your app WebViews.
///   - See https://docs.contentsquare.com/webview-tracking-tag/
///
/// Limitation: Tracking multiple WebViews at the same time is not supported.
///
/// {@endtemplate}
class ContentsquareWebViewTrackerBuilder extends StatefulWidget {
  /// {@template contentsquare.webview.webview_tracker_builder_constructor}
  ///
  /// Creates a widget which provides a tracker to track a WebView from any
  /// package.
  ///
  /// For the tracking to work, you need to:
  ///   - Enable javascript in the WebView.
  ///   - Register the given [ContentsquareWebViewTracker.javascriptChannels]
  ///     to allow the webview to communicate with the SDK.
  ///   - Use [ContentsquareWebViewTracker.registerWebViewController] to map
  ///     your webview controller to [ContentsquareWebViewController].
  ///     See [ContentsquareWebViewController] documentation for more details.
  ///   - Use [ContentsquareWebViewTracker.startPageTracking] to start
  ///     the tracking.
  ///     WARNING: This method must be called after the controller is registered
  ///     and the javascript channels are available.
  ///
  /// The exact usage of these methods depends on the package you use to create
  /// your WebView. Reach out to us if you need help.
  ///
  /// {@endtemplate}
  const ContentsquareWebViewTrackerBuilder({
    Key? key,
    required this.builder,
  }) : super(key: key);

  /// {@template contentsquare.webview.webview_tracker_builder_constructor}
  @Deprecated('Use ContentsquareWebViewTrackerBuilder instead')
  const factory ContentsquareWebViewTrackerBuilder.custom({
    Key? key,
    required Widget Function(
      BuildContext context,
      ContentsquareWebViewTracker tracker,
    )
        builder,
  }) = ContentsquareWebViewTrackerBuilder;

  /// Callback in which the WebView should be built.
  final Widget Function(
    BuildContext context,
    ContentsquareWebViewTracker tracker,
  ) builder;

  @override
  State<ContentsquareWebViewTrackerBuilder> createState() =>
      _a126();
}


class _a126
    extends State<ContentsquareWebViewTrackerBuilder> {
  
  late final a73 = a129(
    a70: a359,
    a69: _a222,
  );

  
  String? _a222() {
    
    
    
    
    if (!mounted) {
      a182.a32(a169());
      return null;
    }

    
    
    
    
    
    
    final a333 = context as Element;

    
    
    
    
    
    
    
    
    
    
    final a334 = a333.renderObject;
    if (a334 == null) {
      a182.a32(a171());
      return null;
    }
    if (a334 is! a128) {
      a182.a28(
        'The type of render object associated with the $ContentsquareWebViewTrackerBuilder widget is ${a334.runtimeType}.\n'
        'This is an error, it should always be a $a128.\n',
      );
      return null;
    }

    
    
    
    
    
    
    
    
    final a335 = a334.a232;
    if (a335 == null) {
      a182.a32(a170());
      return null;
    }

    final a336 = a280(
      a580: a334,
      a581: a131.a238(
        a541: a335.center,
      ),
    );

    return a336.a12(
      a466: (hitElementData) {
        return a188(hitElementData.a65);
      },
      
      
      a465: () => null,
    );
  }

  @override
  void dispose() {
    a73._a229();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _a130(
      a534: _a127(
        a529: widget.builder(context, a73),
        a72: a73.a22,
      ),
    );
  }
}


class _a127
    extends SingleChildRenderObjectWidget {
  const _a127({
    Key? a528,
    required this.a72,
    required Widget a529,
  }) : super(key: a528, child: a529);

  
  final Future<Map<String, dynamic>> Function() a72;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return a128(a71: a72);
  }
}


class a128 extends RenderProxyBox {
  
  a128({
    required this.a71,
  });

  
  final Future<Map<String, dynamic>> Function() a71;
}


class a129 implements ContentsquareWebViewTracker {
  
  a129({
    required this.a70,
    required this.a69,
  });

  
  final a132 a70;

  
  final String? Function() a69;

  
  ContentsquareWebViewController? _a337;

  @override
  void registerWebViewController(ContentsquareWebViewController controller) {
    _a337 = controller;
  }

  @override
  Future<void> startPageTracking() async {
    final _a338 = this._a337;

    if (_a338 == null) {
      return a182.a32(
        a148(
          a576: null,
          a577: 'WebViewController is null, did you forget to use '
              'tracker.onWebViewCreated?',
        ),
      );
    }

    
    final a339 = _a338.isChannelAvailable;
    if (a339 != null) {
      final bool a340;
      try {
        a340 =
            await _a224(
          a531: a339,
        );
      } catch (error) {
        return a182.a32(
          a148(
            a576: await _a338.currentUrl?.call(),
            a577: 'Unexpected error while executing JS code, check the '
                'registration of the javascript channels: $error',
          ),
        );
      }
      if (!a340) {
        return a182.a32(
          a148(
            a576: await _a338.currentUrl?.call(),
            a577:
                'JavascriptChannels are not registered, did you forget to use '
                'tracker.javascriptChannels?',
          ),
        );
      }
    }

    try {
      await _a338.evaluateJavascript(
        '''
${_a223(a530: _a338.javascriptChannelCall)}
        
// This is needed because [_webViewController.evaluateJavascript] only works on
// js code that returns a value (and the bridge registration does not return a 
// value)
true;
''',
      );
    } catch (error) {
      return a182.a32(
        a148(
          a576: await _a338.currentUrl?.call(),
          a577: 'Unexpected error while executing JS code to setup the '
              'Javascript bridge: $error',
        ),
      );
    }

    return a182.a32(
      a153(
        a578: await _a338.currentUrl?.call(),
      ),
    );
  }

  @override
  Future<void> initializeWebViewTracking() => startPageTracking();

  @override
  Set<ContentsquareJavascriptChannel> get javascriptChannels {
    return ContentsquareJavascriptEventEnum.values
        .map(
          (event) => a139(
            a50: Contentsquare(),
            a49: event,
            a48: a69,
          ).a1,
        )
        .toSet();
  }

  
  String _a223({
    required String Function(String channelName, String data)
        a530,
  }) {
    return '''
window.CSJavascriptBridge = {
  
  ${ContentsquareJavascriptEventEnum.values.map(
              (event) => '${event.name}: ${a139(
                a50: Contentsquare(),
                a49: event,
                a48: a69,
              ).a19(
                a472: a530,
              )}',
            ).join(',\n\n').replaceAll('\n', '\n  ')}
};''';
  }

  
  Future<bool> _a224({
    required Future<bool> Function(String) a531,
  }) async {
    for (final a341 in javascriptChannels) {
      final a342 =
          await a531(a341.name);

      if (!a342) {
        return false;
      }
    }

    return true;
  }

  
  
  
  
  var _a343 = IList<Stopwatch>();

  
  Stopwatch _a225() {
    final a344 = Stopwatch();
    _a343 = _a343.add(a344);
    return a344;
  }

  
  void _a226(Stopwatch a532) {
    a532.stop();
    _a343 = _a343.remove(a532);
  }

  
  @visibleForTesting
  static String get a5 => '''
    if (window.cs_wvt == undefined || window.cs_wvt == null) {
      // Cast on 'null' for jsonDecode to decode it to dart null.
      'null';
    } else {
      window.cs_wvt.push(['serializeWebView']);
    }
  ''';

  
  Future<Map<String, dynamic>> a22() async {
    final a345 = _a337;
    if (a345 == null) {
      throw a168();
    }

    
    
    final a346 = _a225();
    a346.start();

    bool a227() {
      return a346.elapsed >=
          a70.a64;
    }

    
    
    
    
    bool a228() {
      return !a346.isRunning;
    }

    Map<String, dynamic>? a347;
    
    
    
    
    while (!(a227() || a228())) {
      final a348 = await a345.evaluateJavascript(
        a5,
      );
      final Object? a349;
      try {
        final a350 = jsonDecode(a348);
        if (a350 is String) {
          
          
          
          a349 = jsonDecode(a350);
        } else {
          a349 = a350;
        }
      } on FormatException catch (e) {
        a182.a32(
          a164(
            a35: a348,
            a34: e,
          ),
        );
        continue;
      }

      
      
      
      
      
      
      
      final a351 = a349 == null ||
          a349 is num ||
          a349 is Map<String, dynamic>;
      if (!a351) {
        a182.a32(
          a163(
            a36: a349,
          ),
        );
      }

      if (a349 is Map<String, dynamic>) {
        a347 = a349;
        break;
      }

      await Future.delayed(
        a70.a63,
      );
    }

    _a226(a346);

    if (a347 == null) {
      throw a165(
        a33: a70.a64,
      );
    }

    return a347;
  }

  
  void _a229() {
    _a343.forEach(_a226);
  }
}


class _a130 extends InheritedWidget {
  const _a130({
    Key? a533,
    required Widget a534,
  }) : super(key: a533, child: a534);

  @override
  bool updateShouldNotify(oldWidget) => false;
}


extension a174 on Element {
  
  bool get _a230 {
    return widget is ContentsquareWebViewTrackerBuilder;
  }

  
  bool get a231 {
    final a352 =
        getElementForInheritedWidgetOfExactType<_a130>();
    return a352 != null || _a230;
  }
}
